# actions classes for specific programs, base classes first:
import unimacro.messagefunctions as mess
import unimacro.natlinkutilsqh as natqh as natqh

class AllActions:
    def __init__(self, progInfo):
        self.reset(progInfo=progInfo)

    # init same as reset:        
    def reset(self, progInfo=None):
        if progInfo is None:
            progInfo = natqh.getProgInfo()
        self.prog, self.topTitle, self.topchild, self.className, self.topHandle = progInfo
        self.progInfo = None
        
    def update(self, newProgInfo=None):
        if newProgInfo is None:
            newProgInfo = natqh.getProgInfo()
        if newProgInfo == self.progInfo:
            return
        # print('allactions: new prog info, overload for your specific program: %s'% self.prog)
        self.progInfo = newProgInfo
        self.prog, self.topTitle, self.topchild, self.className, self.topHandle = newProgInfo
        
    def getCurrentLineNumber(self):
        pass
    
# base actions for applications that are connected through Windows Messages Functions
class MessageActions(AllActions):
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo=progInfo)
        self.classname = None   #  Scintilla can be handled quite generically with setting this variable
        self.prevTabName = None
        self.update(progInfo)

    def update(self, progInfo):
        """if prog info changes, then probably only title changes so notice and do nothing
        """
        if progInfo == self.progInfo:
            return 1 # OK
        self.ctrl = self.handle = self.getInnerHandle(self.topHandle)
        if not self.handle:
            if progInfo and progInfo[2] == 'top':
                print('no handle found for (top) edit control for program: %s'% self.prog)
            return
        self.progInfo = progInfo
        print('updated program info: %s, edit control handle: %s'% (repr(self.progInfo), self.handle)) 
        return self.handle  # None if no valid handle        

    def getCurrentLineNumber(self, handle=None):
        handle = handle or self.handle
        if not handle: return
        linenumber = mess.getLineNumber(handle, classname=self.classname)  # charpos = -1
        return linenumber

    def getNumberOfLines(self, handle=None):
        nl = mess.getNumberOfLines(handle, classname=self.classname)  
        return nl
    
    def isVisible(self, handle=None):
        handle = handle or self.handle
        if not handle: return
        return mess.isVisible(handle, classname=self.classname)
    
    def getEditText(self, handle=None):
        handle = handle or self.handle
        if not handle: return
        tList = mess.getEditText(self.ctrl)
        return ''.join(tList)
    
    getWindowText = getEditText
        
    def replaceSelection(self, output, handle=None):
        """overwrite selection with output
        """
        handle = handle or self.handle
        if not handle: return
        mess.replaceEditText(self.ctrl, output)

    def clearBoth(self, handle=None):
        """clear the target window, and subsequently the dictobj
        """
        handle = handle or self.handle
        if not handle: return
        mess.setEditText(handle, "")
        self.updateState()
        
    def getSelection(self, handle=None):
        """get the selection of the edit control
        
        return a 2 tuple
        """
        handle = handle or self.handle
        if not handle: return
        return mess.getSelection(handle)

    def setSelection(self, start, end, handle=None):
        """change the selection of the edit control
        """
        handle = handle or self.handle
        if not handle: return
        mess.setSelection(handle, start, end)
    
    def getVisibleRegion(self, handle=None):
        """Utility subroutine which calculates the visible region of the edit

         control and returns the start and end of the current visible region.
         Never got this working
        """
        handle = handle or self.handle
        if not handle: return
        return None, None
    
        ## to be investigated
        buf = self.getWindowText()
        return 0, len(buf)
        
        top,bottom,left,right = handle.GetClientRect()
        firstLine = handle.GetFirstVisibleLine()
        visStart = handle.LineIndex(firstLine)

        lineCount = handle.GetLineCount()
        lastLine = lineCount
        for line in range(firstLine+1,lineCount):
            charInLine = handle.LineIndex(line)
            left,top = handle.GetCharPos(charInLine)
            if top >= bottom:
                break
            lastLine = line

        visEnd = handle.LineIndex(lastLine+1)
        if visEnd == -1:
            visEnd = len(handle.GetWindowText())
        return visStart,visEnd
    
    
        


"""
UnittestAutohotkeyactions.py

This module tests the actions, that are performed by autohotkey (autohotkeyactions)

Quintijn Hoogenboom, 2021
"""
import unittest
import time
from pathlib import Path
from profilehooks import profile
from dtactions import autohotkeyactions as ahk
from dtactions.sendkeys import sendkeys

try:
    from dtactions.__init__ import getThisDir, checkDirectory
except ModuleNotFoundError:
    print('\n'.join(['If trying to test this in a git cloned package,',
          'please run this module after "build_package" and "flit install --symlink"',
          'otherwise, this is an unexpected error', 'please report']))
    raise

thisDir = getThisDir(__file__)
dtactionsDir = thisDir.parent

dataDir = Path.home()/".autohotkey"
checkDirectory(dataDir, create=True)

class UnittestAutohotkeyactions(unittest.TestCase):
    """test actions of the module "autohotkeyactions"
    """
    def setUp(self):
        self.notepadHndles = []        
    def tearDown(self):
        for hndle in self.notepadHndles:
            if ahk.SetForegroundWindow(hndle) == 0:
                ahk.killWindow()

    # @profile(filename=dataDir/'profileinfo.txt')
    def tttestSimple(self):
        """only testing GetForegroundWindow and getProgInfo (or GetProgInfo)
        
        Also used for timing the getProgInfo routine
        
        """
        thisHndle = ahk.GetForegroundWindow()
        thisProgInfo = ahk.getProgInfo()
        hndleFromGetProgInfo = thisProgInfo.hndle   # or thisProgInfo[-1]
        mess = '\n'.join(['',
            'GetForegroundWindow and getProgInfo should return the same hndle but:',
            f'getForegroundWindow: {thisHndle}',
            f'from getProgInfo: {hndleFromGetProgInfo}', '='*40])
        self.assertEqual(thisHndle, hndleFromGetProgInfo, mess)
        
        notepadHndle = ahk.ahkBringup("notepad")[-1]
        
        for _ in range(5):
            ahk.SetForegroundWindow(notepadHndle)
            sendkeys('xxx')
            ahk.SetForegroundWindow(thisHndle)
    
        time.sleep(1)
        ahk.killWindow(notepadHndle)
        ahk.SetForegroundWindow(notepadHndle)
        ahk.SetForegroundWindow(thisHndle)
        ahk.SetForegroundWindow(notepadHndle)
        result = ahk.SetForegroundWindow(123)
        mess = '\n'.join(['',
            'SetForegroundWindow should not return hndle of non existing window (123):',
            f'expected: {notepadHndle}',
            f'got {result}', '='*40])
        
        self.assertNotEqual(result, 123, mess)


    def testKillWindow(self):
        """test the autohotkey version of killWindow
        """
        thisHndle = ahk.GetForegroundWindow()
        # thisProgInfo = ahk.getProgInfo()
        
        ## empty notepad window close again:    
        notepadInfo = ahk.ahkBringup("notepad")
        notepadHndle = notepadInfo.hndle
        self.assertTrue(notepadHndle > 0, "notepad should have a valid window hndle, not {notepadHndle}")
        
        ahk.SetForegroundWindow(thisHndle)
        ahk.SetForegroundWindow(notepadHndle)
        ahk.SetForegroundWindow(thisHndle)
        
        
        ## create a child window
        # sendkeys("{ctrl+o}")
        # time.sleep(0.5)
        # childInfo = ahk.getProgInfo()
        # childHndle = childInfo[-1]
        # 
        # print(f'notepad, notepadHndle: {notepadHndle}, child: {childHndle}\n{childInfo}')

        # result = ahk.SetForegroundWindow(thisHndle)        
        ## failed experiment: cannot find back the child window, when you get the top window in the foreground:
        # self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')
        # result = ahk.SetForegroundWindow(notepadHndle)        
        # self.assertTrue(result is True, f'calling notepadHndle seems to succeed, but child window is in front\n\tnotepad: {notepadInfo}\n\tchildwindow: {childHndle}')
        # 
        # sendkeys("{alt+tab}")
        # time.sleep(0.5)
        # sendkeys("{alt+tab}")
        # time.sleep(0.5)
        # 
        # result = ahk.GetForegroundWindow()
        
        # when the hndle of the open child window is given, this works all right
        # the child window is closed, with {esc}, and then the normal killWindow procedure follows.
        # result = ahk.killWindow(childHndle)
        # self.assertTrue(result is True, f'result of killing notepad should be 0, not {result}')

        ## now with text in the window:
        notepadHndle = ahk.ahkBringup("notepad")[-1]
        self.assertTrue(notepadHndle > 0, "notepad should have a valid window hndle, not {notepadHndle}")

        ## print a line of text:
        sendkeys("the quick brown fox...")

        result = ahk.SetForegroundWindow(thisHndle)
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')

        result = ahk.killWindow(notepadHndle)
        self.assertTrue(result is True, f'result of killing notepad should be True, not {result}')
        
        result = ahk.SetForegroundWindow(thisHndle)        
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')

        ####### close with wrong key::::::::::::::::::::::::::::
        ## empty notepad window close again:    
        notepadHndle = ahk.ahkBringup("notepad")[-1]
        self.assertTrue(notepadHndle > 0, "notepad should have a valid window hndle, not {notepadHndle}")
        
        result = ahk.SetForegroundWindow(thisHndle)
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')

        result = ahk.killWindow(notepadHndle, key_close="{alt+f5}")   # should be {alt+f4}
        self.assertFalse(result is True, f'result of killing notepad should NOT be True\n\t{result}')

        result = ahk.killWindow(notepadHndle)   # should be {alt+f4}
        self.assertTrue(result is True, f'result of killing notepad should be True, not\n\t{result}')
        
        result = ahk.SetForegroundWindow(thisHndle)        
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')
        
        ## now with text in the window:
        notepadHndle = ahk.ahkBringup("notepad")[-1]
        self.assertTrue(notepadHndle > 0, "notepad should have a valid window hndle, not {notepadHndle}")

        ## print a line of text:
        sendkeys("the quick brown fox...")

        result = ahk.SetForegroundWindow(thisHndle)
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')

        result = ahk.killWindow(notepadHndle, key_close_dialog="{alt+m}")
        self.assertFalse(result is True, f'result of killing notepad should be not True\n\t{result}')

        ## now do it good:
        result = ahk.killWindow()
        self.assertTrue(result is True, f'result of killing notepad should be True, not:\n\t{result}')

        result = ahk.SetForegroundWindow(thisHndle)        
        self.assertTrue(result is True, f'calling window should be in the foreground again {thisHndle}')

    def tttestAhkBringupNotePad(self):
        """test easy bringup of Notepad

        second call tries to activate to the first openend instance        
        """
        ### test notepad:
        thisHndle = ahk.GetForegroundWindow()
        notepadInfo = ahk.ahkBringup("notepad")
        notepadHndle = notepadInfo.hndle
        self.assertGreater(notepadHndle, 0, "notepad should have a valid window hndle, not {notepadHndle}")
        self.assertEqual(notepadInfo.prog, 'notepad', 'notepadInfo, prog does not match')
        notepadTitle = notepadInfo.title
        searchTitle = notepadTitle.split()[0]   ## only Naamloos
        
        # go back, then pick up window with same title:
        ahk.SetForegroundWindow(thisHndle)
        notepadInfo2 = ahk.ahkBringup("notepad", title=searchTitle)
        notepadHndle2 = notepadInfo2.hndle
        self.assertGreater(notepadHndle2, 0, "notepad 2 should have a valid window hndle, not {notepadHndle2}")
        self.assertEqual(notepadHndle2, notepadHndle, "notepad 2 should be same window as notepad")
        self.assertTupleEqual(notepadInfo2, notepadInfo2, 'notepadInfo2 should equal notepadInfo')
        
        ahk.killWindow(notepadHndle)
        ahk.killWindow(notepadHndle2)
        nowThisHndle = ahk.GetForegroundWindow()
        self.assertEqual(nowThisHndle, thisHndle, 'test should end up in same window as it started')

        ## open a filepath, the errormessagefromahk.txt file (1 empty line)
        filepath = Path.home()/'.autohotkey'/'errormessagefromahk.txt'
        self.assertTrue(filepath.is_file(), f'file {filepath} should exist')
        notepadInfo3 = ahk.ahkBringup("notepad", filepath=filepath)
        notepadHndle3 = notepadInfo3.hndle
        self.assertGreater(notepadHndle3, 0, "notepad 3 should have a valid window hndle, not {notepadHndle2}")
        self.assertEqual(notepadInfo3.title, str(filepath), f'filepath should be in window title: {filepath}')



    def tttestAhkBringupThunderbirdAppTitle(self):
        """test starting or bringing to the foreground Thunderbird
        
        Note the default waitForStart is increased from 1 to 3 seconds, as starting Thunderbird takes
        a bit of time.
        """
        ### test Thunderbird, app and title, activate if already open...
        thisHndle = ahk.GetForegroundWindow()
        ## if thunderbird is not open, and you remove waitForStart, you probably get the fail message below,
        ## as it takes more than a second to start thunderbird...
        thunderbirdInfo = ahk.ahkBringup("thunderbird", title="Mozilla Thunderbird", waitForStart=5)
        if isinstance(thunderbirdInfo, str):
            self.fail(f'starting/activating thunderbird failed, message: {thunderbirdInfo}')
        thunderbirdHndle = thunderbirdInfo.hndle
        
        ahk.SetForegroundWindow(thisHndle)
      
        self.assertGreater(thunderbirdHndle, 0, "thunderbird should have a valid window hndle, not {thunderbirdHndle}")
        self.assertEqual(thunderbirdInfo.prog, 'thunderbird', 'thunderbirdInfo, prog does not match')

        # leave thunderbird open after testing
        # ahk.killWindow(thunderbirdHndle)

        ahk.SetForegroundWindow(thisHndle)
        nowThisHndle = ahk.GetForegroundWindow()
        self.assertEqual(nowThisHndle, thisHndle, 'test should end up in same window as it started')
        
        ## switch to thunderbird twice, and then back:
        thunderbirdInfoTwo = ahk.ahkBringup("thunderbird", title="Mozilla Thunderbird", waitForStart=5)
        if isinstance(thunderbirdInfoTwo, str):
            self.fail(f'starting/activating thunderbird failed, message: {thunderbirdInfoTwo}')
        thunderbirdInfoThree = ahk.ahkBringup("thunderbird", title="Mozilla Thunderbird", waitForStart=5)
        if isinstance(thunderbirdInfoThree, str):
            self.fail(f'starting/activating thunderbird failed, message: {thunderbirdInfoThree}')

        ## back to calling window:        
        ahk.SetForegroundWindow(thisHndle)
        nowThisHndle = ahk.GetForegroundWindow()
        self.assertEqual(nowThisHndle, thisHndle, 'test should end up in same window as it started')

        

    def tttestAhkBringupWindword(self):
        """test starting or bringing to the foreground Windword
        
        """
        ### test winword, app and title, activate if already open...
        thisHndle = ahk.GetForegroundWindow()
        ## if winword is not open, and you remove waitForStart, you probably get the fail message below,
        ## as it takes more than a second to start winword...
        winwordInfo = ahk.ahkBringup("winword", title="Word", waitForStart=10)
        if isinstance(winwordInfo, str):
            self.fail(f'starting winword failed, message: "{winwordInfo}"')
        winwordHndle = winwordInfo.hndle
        
        ahk.SetForegroundWindow(thisHndle)
      
        self.assertGreater(winwordHndle, 0, "winword should have a valid window hndle, not {winwordHndle}")
        self.assertEqual(winwordInfo.prog, 'WINWORD', 'winwordInfo, prog does not match')

        # leave winword open after testing
        ahk.SetForegroundWindow(thisHndle)
        nowThisHndle = ahk.GetForegroundWindow()
        self.assertEqual(nowThisHndle, thisHndle, 'test should end up in same window as it started')

def run():
    """run the unittest procedure"""
    print('starting UnittestAutohotkeyactions')

    suite = unittest.makeSuite(UnittestAutohotkeyactions, 'test')
    unittest.TextTestRunner().run(suite)

if __name__ == "__main__":
    run()



"""actions from application komodo

see www.activestate.com. Does not work (yet)
"""
import win32gui
import ctypes
import win32api
import unimacro.messagefunctions as mf
from .actionbases import MessageActions
import pprint

class KomodoActions(MessageActions):
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        self.classname = "Scintilla"
        
    def getInnerHandle(self, handle):
        # cannot get the correct "inner handle"
        return
        nextHndle = handle 
        user32 = ctypes.windll.user32
        #
        controls = mf.findControls(handle, wantedClass="Scintilla")
        print('Scintilla controls: %s'% controls)
        for c in controls:
            ln = self.getCurrentLineNumber(c)
            numberLines = self.getNumberOfLines(c)
            visible1 = self.isVisible(c)
            info = win32gui.GetWindowPlacement(c) 
            print('c: %s, linenumber: %s, nl: %s, info: %s'% (c, ln, numberLines, repr(info)))
            parent = c
            while 1:
                parent = win32gui.GetParent(parent)
                clName = win32gui.GetClassName(parent)
                visible = self.isVisible(parent)
                info = win32gui.GetWindowPlacement(parent) 
                print('parent: %s, class: %s, visible: %s, info: %s'% (parent, clName, visible, repr(info)))
                if parent == handle:
                    print('at top')
                    break
        

# several tries for getting introspection of the Komodo controls failed sofar (QH, October 2013) 
import ctypes
import win32con
def get_windows(startWith=None):
    '''Returns windows in z-order (top first)'''
    user32 = ctypes.windll.user32
    lst = []
    top = user32.GetTopWindow(startWith)
    if not top:
        return lst
    lst.append(top)
    next = top
    print('top: %s'% next)
    while True:
        
        next = user32.GetTopWindow(next)
        print('next: %s'% next)
        if not next:
            break
    #    lst.append(next)
    #return lst        

if __name__ == '__main__':

    focus =66284
    
    print('focus; %s'% focus)
    progInfo = ('komodo', 'babbababa', 'top', focus)
    ka = KomodoActions( progInfo )
    get_windows()
    #tw = mf.findTopWindows(wantedText="komodo")
    #for t in tw:
    #    print 't:%s'% t
    #    lst = get_windows(t)
    #    for l in lst:
    #        visible = win32gui.SendMessage(l, win32con.WS_VISIBLE)
    #        print 'l: %s, visible: %s'% (l, visible)
    ##    #pr
    ##    topchild = user32.GetTopWindow(t)
    ##    nextchild = user32.GetTopWindow(topchild)
    ##    print 'tophandle: %s, topchild: %s, nextchild: %s'% (t, topchild, nextchild)
    ##    
    #    pprint.pprint( mf.dumpWindow(t) )
    #    print '----------------------------------------------------------'        
    #    


"""actions from application komodo

see www.activestate.com. Does not work yet for s&s or line numbers modulo 100
now special (QH) metaactions for very special tasks on selections of a file only
"""
import win32gui
import ctypes
import win32api
import unimacro.messagefunctions as mf
from .actionbases import AllActions
from unimacro.actions import doAction as action
from unimacro.actions import doKeystroke as keystroke
import natlinkclipboard
import time


# class KomodoActions(MessageActions):
class KomodoActions(AllActions):
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo)
        self.classname = "Komodo"
        
    def getInnerHandle(self, handle):
        # cannot get the correct "inner handle"
        return
        nextHndle = handle 
        user32 = ctypes.windll.user32
        #
        controls = mf.findControls(handle, wantedClass="Scintilla")
        print('Scintilla controls: %s'% controls)
        for c in controls:
            ln = self.getCurrentLineNumber(c)
            numberLines = self.getNumberOfLines(c)
            visible1 = self.isVisible(c)
            info = win32gui.GetWindowPlacement(c) 
            print('c: %s, linenumber: %s, nl: %s, info: %s'% (c, ln, numberLines, repr(info)))
            parent = c
            while 1:
                parent = win32gui.GetParent(parent)
                clName = win32gui.GetClassName(parent)
                visible = self.isVisible(parent)
                info = win32gui.GetWindowPlacement(parent) 
                print('parent: %s, class: %s, visible: %s, info: %s'% (parent, clName, visible, repr(info)))
                if parent == handle:
                    print('at top')
                    break
    def getCurrentLineNumber(self, handle=None):
        debug = 0
        t1 = time.time()
        if self.topchild == "child":
            return 0
        cb = natlinkclipboard.Clipboard(save_clear=True, debug=debug)  # clear "debug" to get rid of timing line
        shortcutkey = "{alt+shift+c}"
        keystroke(shortcutkey)
        
        # # now collect the clipboard, at most waiting 10 intervals of 0.1 second.
        result = cb.get_text(waiting_interval=0.01, waiting_iterations=10)    # should be the current line number
        # print(f'result from clipboard: {result}')
        t2 = time.time()
        lapse = t2 - t1
        if debug:
            print(f'time in getCurrentLineNumber: {lapse:.3f}')
        try:
            return int(result)
        except (ValueError, TypeError):
            return 0
        
    def metaaction_makeunicodestrings(self, dummy=None):
        """for doctest testing, put u in front of every string
        """
        print('metaaction_makeunicodestrings, for Komodo')
        natut.playString("{ctrl+c}")
        t = natqh.getClipboard()
        print('in: %s'% t)
        t = replaceStringToUnicode(t)
        natqh.setClipboard(t)
        natut.playString("{ctrl+v}{down}")

def _test():
    """do doctests, for changing function
    """
    import doctest
    doctest.testmod()
    
